/******************************************/
/* Wmseti - Mini SETI@Home monitor        */
/******************************************/

/******************************************/
/* This program is free software; you can redistribute it and/or
/* modify it under the terms of the GNU General Public License
/* as published by the Free Software Foundation; either version 2
/* of the License, or (at your option) any later version.
/* 
/* This program is distributed in the hope that it will be useful,
/* but WITHOUT ANY WARRANTY; without even the implied warranty of
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/* GNU General Public License for more details.
/* 
/* You should have received a copy of the GNU General Public License
/* along with this program; if not, write to the Free Software
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
/******************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <pwd.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#ifdef __sun__
#include <procfs.h>
#endif

#include "wmgeneral.h"
#include "misc.h"

#if GRAPHIC == 1
#include "xpm/wmseti-brownwhite.xpm"
#elif GRAPHIC == 2
#include "xpm/wmseti-greyblue.xpm"
#elif GRAPHIC == 3
#include "xpm/wmseti-brownwhitegreyblue.xpm"
#elif GRAPHIC == 4
#include "xpm/wmseti-custom.xpm"
#else
#undef GRAPHIC
#include "xpm/wmseti-default.xpm"
#include "xpm/wmseti-lcd.xpm"
#endif

#define CHAR_X		4		/* Char X origin			*/
#define CHAR_Y		6		/* Char Y origin			*/
#define CHAR_WIDTH	6		/* Char width in pixels			*/
#define CHAR_HEIGHT	7		/* Char height in pixels		*/
#define PIXMAP_X	3		/* Pixmap X origin			*/
#define PIXMAP_Y	4		/* Pixmap Y origin			*/
#define PIXMAP_WIDTH	58		/* Pixmap width in pixels		*/
#define PIXMAP_HEIGHT	56		/* Pixmap height in pixels		*/
#define LINE_NUMBER	5		/* Number of lines			*/
#define LINE_LENGTH	9		/* Characters per line			*/
#define LINE_WIDTH	54		/* Line width in pixels			*/
#define LINE_HEIGHT	11		/* Line height in pixels		*/
#define MASK_WIDTH	64		/* Mask width in pixels			*/
#define MASK_HEIGHT	64		/* Mask height in pixels		*/

#define SETI_DIRMAX	10

#define STAT_TITLE	1
#define STAT_PROG	2
#define STAT_CPU	3
#define STAT_TIMELEFT	4
#define STAT_NWUS	5
#define STAT_NRESULTS	6
#define STAT_PID	7
#define STAT_TOTALCPU	8
#define STAT_BAR	9

#define DISP_INFO_A	0		/* Show info lines 1-5			*/
#define DISP_INFO_B	1		/* Show info lines 6-10			*/
#define DISP_INFO_CLEAR	2		/* Reset display to info		*/
#define DISP_ALIEN_DIR	3		/* Show alien with directory indicator	*/
#define DISP_ALIEN	4		/* Show alien				*/

#define FILE_USERINFO	"user_info.sah"	/* User statistics			*/
#define FILE_STATE	"state.sah"	/* State of setiathome workunit		*/
#define FILE_PID	"pid.sah"	/* Pid of setiathome workunit		*/
#define FILE_WTEMP	"wtemp.sah"	/* Exists if downloading new workunit	*/
#define FILE_RESULT	"result.sah"	/* Exists if finished workunit		*/

/******************************************/
/* Structures                             */
/******************************************/

struct {
	char **pm;
#ifdef GRAPHIC
	int face;
	int bar;
	int numface;
	int perface;
#else
	char *description;
#endif
#if GRAPHIC == 1
} pixmap = { wmseti_brownwhite_xpm, 0, 1, 2, 3 };
#elif GRAPHIC == 2
} pixmap = { wmseti_greyblue_xpm, 0, 1, 2, 3 };
#elif GRAPHIC == 3
} pixmap = { wmseti_brownwhitegreyblue_xpm, 0, 1, 4, 3 };
#elif GRAPHIC == 4
} pixmap = { wmseti_custom_xpm, 1, 0, 2, 4 };
#else
} pixmap[] = {
        { wmseti_default_xpm, "Default" },
        { wmseti_lcd_xpm, "Lcd" },
};
#endif

struct {
	pid_t pid;
	char status;
	char progress[LINE_LENGTH];
	char cputime[LINE_LENGTH];
	char timeleft[LINE_LENGTH];
	int nwus;
	int nresults;
	char totalcpu[LINE_LENGTH];
} stats;

struct {
	char program[128];
	char dir[SETI_DIRMAX][128];
        int dirtotal;
	int dircurrent;
	int display;
        int line[2*LINE_NUMBER];
        int update;
        int refresh;
} seti;

/******************************************/
/* Globals                                */
/******************************************/

char mask[MASK_WIDTH * MASK_HEIGHT];

int debug = 0;
#ifndef GRAPHIC
int theme = 0;
int nthemes = 2;
#endif

/******************************************/
/* Functions                              */
/******************************************/

void read_config(int, char **);
int insert_setidir(const char *);
void usage();
int expand_dir(char *);
int existing_dir(const char *);
int print_stats(void);
void format_string(char *, char *);
void blit_bar(int, int, int, int, int, int, int);
void blit_string(char *, int, int);
#ifdef GRAPHIC
int print_alien(void);
#endif
void read_stats(const char *);
pid_t get_pid(const char *);
void read_workunit_stats(const char *);
void read_user_stats(const char *);
char *secs_2_time(const double, char *);
char *secs_2_epoch(const double, char *);
char get_status(const pid_t);

/******************************************/
/* Main                                   */
/******************************************/

int main(int argc, char **argv)
{
	XEvent Event;

	int dirseti = 0;
	int pauseseti = 0;
	int runseti = 0;
	int count = 0;

	read_config( argc, argv );

	count = seti.update;

#ifdef GRAPHIC
	createXBMfromXPM(mask, pixmap.pm, MASK_WIDTH, MASK_HEIGHT);
	openXwindow(argc, argv, pixmap.pm, mask, MASK_WIDTH, MASK_HEIGHT);
#else
	createXBMfromXPM(mask, pixmap[theme].pm, MASK_WIDTH, MASK_HEIGHT);
	openXwindow(argc, argv, pixmap[theme].pm, mask, MASK_WIDTH, MASK_HEIGHT);
#endif

	while (1) {
		waitpid(0, NULL, WNOHANG);

		if ( count >= seti.update ) {
			count = 0;
			read_stats( seti.dir[seti.dircurrent] );

#ifdef GRAPHIC
			if ( seti.display < DISP_ALIEN_DIR ) {
				if ( seti.display == DISP_INFO_CLEAR )
					copyXPMArea(0, 111, PIXMAP_WIDTH, PIXMAP_HEIGHT, PIXMAP_X, PIXMAP_Y);
				seti.display = ! seti.display;
				print_stats();
			} else {
				print_alien();
				if ( seti.display == DISP_ALIEN_DIR )
					seti.display = DISP_ALIEN;
			}
#else
			print_stats();
			seti.display = ! seti.display;
#endif

			RedrawWindow();
			count = 0;
		}

		while ( XPending(display) ) {
			XNextEvent(display, &Event);
			switch ( Event.type ) {
				case Expose:
					RedrawWindow();
					break;
				case DestroyNotify:
					XCloseDisplay(display);
					exit(0);
				case ButtonRelease:
#ifdef GRAPHIC
					if ( Event.xbutton.button == 1 ) {
						if ( Event.xbutton.x < MASK_WIDTH / 2 && Event.xbutton.y < MASK_HEIGHT / 2 ) {
							if ( seti.display < DISP_ALIEN_DIR ) {
								if ( seti.dirtotal == 1 ) {
									pixmap.face++;
									seti.display = DISP_ALIEN;
								} else
									seti.display = DISP_ALIEN_DIR;
							} else
								seti.display = DISP_INFO_CLEAR;
						} else if ( Event.xbutton.x < MASK_WIDTH / 2  && seti.display > DISP_INFO_CLEAR )
							pixmap.bar = ! pixmap.bar;
						else if ( seti.dirtotal > 1 )
							dirseti = 1;
					}
#else
					if ( Event.xbutton.button == 1 && seti.dirtotal > 1 )
						dirseti = 1;
#endif
					if ( Event.xbutton.button == 2 )
						runseti = 1;
					if ( Event.xbutton.button == 3 )
						pauseseti = 1;
					if ( Event.xbutton.button == 4 && seti.dirtotal > 1 )
						dirseti = 1;
					if ( Event.xbutton.button == 5 && seti.dirtotal > 1 )
						dirseti = -1;
					read_stats( seti.dir[seti.dircurrent] );
					count = seti.update;
					break;
			}
		}

		if ( pauseseti ) {
			if ( stats.status != 'U') {
				if ( stats.status == 'R' )
					kill( stats.pid, SIGSTOP );
				else
					kill( stats.pid, SIGCONT );
			}

			pauseseti = 0;
		}

		if ( runseti ) {
			if ( stats.status == 'U' ) {
				chdir( seti.dir[seti.dircurrent] );
				execCommand( seti.program );
			} else {
				if ( stats.status != 'R' )
					kill( stats.pid, SIGCONT );
				kill( stats.pid, SIGTERM );
			}

			runseti = 0;
		}

		if ( dirseti ) {
			if ( dirseti > 0 ) {
				seti.dircurrent++;
				if ( seti.dircurrent >= seti.dirtotal )
					seti.dircurrent = 0;
			} else {
				seti.dircurrent--;
				if ( seti.dircurrent < 0 )
					seti.dircurrent = seti.dirtotal - 1;
			}

#ifdef GRAPHIC
			pixmap.face++;
			if ( seti.display > DISP_INFO_CLEAR )
				seti.display = DISP_ALIEN_DIR;
#endif
			dirseti = 0;
		}

		usleep(seti.refresh);
		count = count + seti.refresh;
	}
}

/******************************************/
/* Read config file                       */
/******************************************/

void read_config(int argc, char **argv)
{
	int i, cmddir = 0;
	char buf[256];
	char left[128], right[128];
	char *c;
	FILE *fp;

	seti.dirtotal = 0;
	seti.dircurrent = 0;
	seti.display = 0;
	seti.update = 2500000;
	seti.refresh = 100000;
	strcpy( seti.program, "setiathome" );

	for ( i = 0; i < SETI_DIRMAX; i++ )
		seti.dir[i][0] = 0;

	seti.line[0] = STAT_TITLE;
	seti.line[1] = STAT_PROG;
	seti.line[2] = STAT_CPU;
	seti.line[3] = STAT_TIMELEFT;
	seti.line[4] = STAT_NRESULTS;
	seti.line[5] = STAT_TITLE;
	seti.line[6] = STAT_PROG;
	seti.line[7] = STAT_CPU;
	seti.line[8] = STAT_BAR;
	seti.line[9] = STAT_TOTALCPU;

	while ( ( i = getopt(argc, argv, "hvs:r:d:p:abt:")) != -1 ) {
		switch ( i ) {
			case 'v':
				debug = 1;
				break;
			case 'h':
				usage();
				exit(1);
		}
	}
	optind = 0;

	strcpy( buf,  "~/.wmsetirc" );
	expand_dir( buf );

	if ( fp = fopen( buf, "r" ) ) {
		while ( fgets( buf, sizeof(buf), fp ) ) {
			if ( sscanf( buf, "%s %s", &left, &right ) == 2 ) {
				if ( atoi(left) > 0 && atoi(left) <= 10 ) {
					i = atoi(left) - 1;
					if ( ! strcasecmp( right, "title" ) )
						seti.line[i] = STAT_TITLE;
					else if ( ! strcasecmp( right, "prog" ) )
						seti.line[i] = STAT_PROG;
					else if ( ! strcasecmp( right, "cpu" ) )
						seti.line[i] = STAT_CPU;
					else if ( ! strcasecmp( right, "timeleft" ) )
						seti.line[i] = STAT_TIMELEFT;
					else if ( ! strcasecmp( right, "nwus" ) )
						seti.line[i] = STAT_NWUS;
					else if ( ! strcasecmp( right, "nresults" ) )
						seti.line[i] = STAT_NRESULTS;
					else if ( ! strcasecmp( right, "pid" ) )
						seti.line[i] = STAT_PID;
					else if ( ! strcasecmp( right, "totalcpu" ) )
						seti.line[i] = STAT_TOTALCPU;
					else if ( ! strcasecmp( right, "bar" ) )
						seti.line[i] = STAT_BAR;
				} else if ( ! strcasecmp( left, "setidir" ) ) {
					if ( c = strchr( right, '\n' ) )
						*c = 0;
					insert_setidir( right );
				} else if ( ! strcasecmp( left, "setiprg" ) ) {
					if ( c = strchr( buf, '\n' ) )
						*c = 0;
					strcpy( seti.program, strstr( buf, right ) );
				} else if ( ! strcasecmp( left, "refresh" ) )
					seti.refresh = atoi(right) * 1000;
				else if ( ! strcasecmp( left, "update" ) )
					seti.update = atoi(right) * 1000;
			}
		}
		fclose(fp);
	}

	while ( ( i = getopt(argc, argv, "hvs:r:d:p:abt:")) != -1 ) {
		switch ( i ) {
			case 's':
				if ( optarg )
					seti.update = atoi(optarg) * 1000;
				break;
			case 'r':
				if ( optarg )
					seti.refresh = atoi(optarg) * 1000;
				break;
			case 'd':
				if (optarg) {
					if ( cmddir == 0 ) {
						seti.dirtotal = 0;
						cmddir = 1;
					}
					insert_setidir( optarg );
				}
				break;
			case 'p':
				if (optarg)
					strcpy( seti.program, optarg );
				break;
#ifdef GRAPHIC
			case 'a':
				seti.display = DISP_ALIEN_DIR;
				break;
			case 'b':
				pixmap.bar = 0;
				break;
#else
			case 't':
				if (optarg) {
					i = atoi(optarg);
					if ( i > 0 && i <= nthemes ) {
						theme = i - 1;
						break;
					}
				}
#endif
			case 'h':
			case 'v':
				break;
			default:
				usage();
				exit(1);
		}
	}

	expand_dir( seti.program );

	if ( ! seti.dir[0][0] ) {
		if ( ! insert_setidir( "~/setiathome" ) ) {
			fprintf(stderr, "DIR: Default directory failed, no setiathome\n");
			exit(0);
		}
	}

        if ( debug ) {
		fprintf(stderr, "P: %s\n", seti.program );
		for ( i = 0; i < seti.dirtotal; i++ )
			fprintf(stderr, "D%d: %s\n", i, seti.dir[i] );
		for ( i = 0; i < 10; i++ )
			fprintf(stderr, "S%d: %d\n", i, seti.line[i] );
        }


}

/******************************************/
/* Usage                                  */
/******************************************/

void usage(void)
{
#ifndef GRAPHIC
        int i;
#endif

        fprintf(stderr, "\nWmseti - SETI@Home Monitor Dock V%s\n\n", VERSION);
        fprintf(stderr, "Usage: wmseti [ options... ]\n\n");
        fprintf(stderr, "\t-display <display name>\n");
        fprintf(stderr, "\t-geometry +XPOS+YPOS\tinitial window postion\n");
	fprintf(stderr, "\t-s <...>\t\tsample rate in milliseconds  (default:%d)\n", seti.update / 1000 );
	fprintf(stderr, "\t-r <...>\t\trefresh rate in milliseconds (default:%d)\n", seti.refresh / 1000 );
        fprintf(stderr, "\t-d [...]\t\tspecify a SETI@home directory (overrides config)\n");
	fprintf(stderr, "\t\t\t\tcan be applied multiple times: -d [...] -d [...]\n");
        fprintf(stderr, "\t-p [...]\t\tspecify command line options for setiathome\n");
        fprintf(stderr, "\t-v\t\t\tdebug\n");
#ifdef GRAPHIC
        fprintf(stderr, "\t-a\t\t\tshow alien face\n");
        fprintf(stderr, "\t-b\t\t\thide progress bar on alien face\n");
#else
        fprintf(stderr, "\t-t [1..%d]\t\tchanges the theme pixmap\n", nthemes);
        for (i = 0; i < nthemes; ++i)
                fprintf(stderr, "\t\t\t\t%d - %s\n", i+1, pixmap[i].description);
#endif
	fprintf(stderr, "\t-h\t\t\tprints this help\n");
}

/******************************************/
/* Check and add seti@home directory      */
/******************************************/

int insert_setidir(const char *dir)
{
	int i;
	char buf[128];
	DIR *dp;
	FILE *fp;

	if ( seti.dirtotal >= SETI_DIRMAX ) {
		if ( debug )
			fprintf(stderr, "DIR: Too many directories\n");
		return 0;
	}

	strcpy( buf, dir );

	if ( ! expand_dir(buf) ) {
		if ( debug )
			fprintf(stderr, "DIR: Failed to expand directory - %s\n", buf);
		return 0;
	}

	if ( dp = opendir(buf) )
		closedir(dp);
	else {
		if ( debug )
			fprintf(stderr, "DIR: Failed to access directory - %s\n", buf);
		return 0;
	}

	if ( seti.dirtotal ) {
		for ( i = 0; i < seti.dirtotal; i++ ) {
			if ( ! strcmp( buf, seti.dir[i] ) ) {
				if ( debug )
					fprintf(stderr, "DIR: Duplicate directory - %s\n", buf);
				return 0;
			}
		}
	}

	strcpy( seti.dir[seti.dirtotal], buf );

	seti.dirtotal++;

	return 1;
}

/******************************************/
/* Expand directory                       */
/******************************************/

int expand_dir(char *dir)
{
	char rootdir[128], subdir[128];
	struct passwd pswd;
	struct passwd *pw = &pswd;
	int i = 0;

	/* Directory is absolute 				*/

	if ( *dir == '/' )
		return 1;

	/* Split into relevant components			*/

	while ( *(dir+i) != '/' && *(dir + i) != '\0')
		i++;
	strcpy( subdir, dir+i );

	/* Expand "~/" to home directory OR			*/
	/* Expand "~other/" to other users home directory	*/

	if ( *dir == '~' && *(dir+1) == '/' ) {
		if ( ! strcpy( rootdir, getenv("HOME") ) ) {
			if ( pw = getpwuid( getuid() ) )
				strcpy( rootdir, pw->pw_dir );
			else
				return 0;
		}
	} else if ( *dir == '~' && *(dir+1) != '/' ) {
		i = 1;
		while ( *(dir+i) != '/' ) {
			rootdir[i-1] = *(dir+i);
			i++;
		}
		rootdir[i-1] = '\0';

		if ( pw = getpwnam(rootdir) )
			strcpy( rootdir, pw->pw_dir );
		else
			return 0;
	} else
		return 0;

	strcat( rootdir, subdir);
	strcpy( dir, rootdir );

	return 1;
}

/******************************************/
/* Draw statistics                        */
/******************************************/

int print_stats(void)
{
	int i, j;
	int ycoord; 
	char left[10], right[10];
	
	for ( i = 0; i < LINE_NUMBER; i++ ) {
		ycoord = CHAR_Y + LINE_HEIGHT * i;

		j = seti.line[i + 5 * seti.display];

		strcpy( right, "\0" );

		if ( j == STAT_BAR )
			blit_bar(0, 97, LINE_WIDTH - 1, CHAR_HEIGHT, atoi(stats.progress), CHAR_X + 1, ycoord);
		else {
			if (  j == STAT_TITLE ) {
				strcpy( left, "wmseti" );
				if ( seti.dirtotal > 1 )
					sprintf( right, "%c", 65 + seti.dircurrent );
			} else if ( j == STAT_PID ) {
				strcpy( left, "PID:" );
				sprintf( right, "%i", stats.pid );
			} else if ( j == STAT_PROG ) {
				strcpy( left, stats.progress );
				sprintf( right, "%c", stats.status );
			} else if ( j == STAT_CPU ) {
				strcpy( left, stats.cputime );
			} else if ( j == STAT_TIMELEFT ) {
				strcpy( left, stats.timeleft );
			} else if ( j == STAT_NWUS ) {
				strcpy( left, "Rx:" );
				sprintf( right, "%i", stats.nwus );
			} else if ( j == STAT_NRESULTS ) {
				strcpy( left, "Tx:" );
				sprintf( right, "%i", stats.nresults );
			} else if ( j == STAT_PID ) {
				strcpy( left, "PID:" );
				sprintf( right, "%i", stats.pid );
			} else if ( j == STAT_TOTALCPU ) {
				strcpy( left, "Pt:" );
				strcpy( right, stats.totalcpu );
			}

			format_string( left, right );
			blit_string( left, CHAR_X, ycoord );
		}
	}
}

/******************************************/
/* Format string                          */
/******************************************/

void format_string(char *left, char *right)
{
	int n = strlen(left) + strlen(right);

	if ( n <= LINE_LENGTH ) {
		strncat(left, "         ", LINE_LENGTH - n );
		strcat(left, right);
	} else
		strcpy(left, "  ERROR  ");
}

/******************************************/
/* Blit string at co-ordinates            */
/******************************************/

void blit_string(char *name, int x, int y)
{
	int i;
	int c;
	int k;

	k = x;

	for ( i = 0; name[i]; i++ ) {
		c = toupper( name[i] );
		if ( c >= '0' && c <= '9' ) {
			c -= '0';
			copyXPMArea( c * 6, 65, CHAR_WIDTH, CHAR_HEIGHT, k, y ); 
		} else if ( c >= 'A' && c <= 'J' ) {   
			c -= 'A';
			copyXPMArea( c * 6, 73, CHAR_WIDTH, CHAR_HEIGHT, k, y );
		} else if ( c >= 'K' && c <= 'T' ) {
			c -= 'K';
			copyXPMArea( c * 6, 81, CHAR_WIDTH, CHAR_HEIGHT, k, y );
		} else if ( c >= 'U' && c <= 'Z' ) {
			c -= 'U';
			copyXPMArea( c * 6, 89, CHAR_WIDTH, CHAR_HEIGHT, k, y );
		} else if ( c == ':' )
			copyXPMArea( 36, 89, CHAR_WIDTH, CHAR_HEIGHT, k, y );
		else if ( c == '.' )
			copyXPMArea( 42, 89, CHAR_WIDTH, CHAR_HEIGHT, k, y);
		else if ( c == '/' )
			copyXPMArea( 48, 89, CHAR_WIDTH, CHAR_HEIGHT, k, y);
		else if ( c == '%' )
			copyXPMArea( 54, 89, CHAR_WIDTH, CHAR_HEIGHT, k, y);
		else
			copyXPMArea( 54, 97, CHAR_WIDTH, CHAR_HEIGHT, k, y );
	k += 6;
	}
}

/******************************************/
/* Blit percentage bar at co-ordinates    */
/******************************************/

void blit_bar(int sx, int sy, int w, int h, int percent, int dx, int dy)
{
	int tx;

	if ( percent <= 100 )
		tx = w * percent / 100;
	else
		tx = w;

	if ( tx > 0 )
		copyXPMArea(sx, sy ,tx, h, dx, dy);
	if ( tx < w )
		copyXPMArea(sx+tx, sy+h, w-tx, h, dx+tx, dy);
}

/******************************************/
/* Draw alien face                        */
/******************************************/

#ifdef GRAPHIC
int print_alien(void)
{
	int sx = 64;
	int sy = 0;
	int face = 0;
	int barcolour = 0;
	int i;

	if ( pixmap.face >= pixmap.numface )
		pixmap.face = 0;

	face = pixmap.perface * ( 100 - atoi(stats.progress) ) / 100;
	if ( face >= pixmap.perface )
		face = pixmap.perface - 1;

	sx += pixmap.face * PIXMAP_WIDTH;
	sy = face * PIXMAP_HEIGHT;

	copyXPMArea(sx, sy, PIXMAP_WIDTH, PIXMAP_HEIGHT, PIXMAP_X, PIXMAP_Y);

	if ( pixmap.bar ) {
		if ( stats.status == 'U' )
			barcolour += 2;
		else if ( stats.status != 'R' )
			barcolour += 1;

		copyXPMArea(0, 167 + barcolour, atoi(stats.progress) * LINE_WIDTH / 100, 1, CHAR_X + 1, 56);
	}

	if ( seti.display == DISP_ALIEN_DIR ) {
		for ( i = 0; i <= seti.dircurrent; i++ )
			copyXPMArea(55, 167, 3, 3, PIXMAP_X + 3, PIXMAP_Y + 3 + i * 5);
	}
}
#endif

/******************************************/
/* Read statistics                        */
/******************************************/

void read_stats(const char *dir)
{
	stats.pid = 0;
	stats.status = 'U';
	strcpy( stats.progress, "00.00%" );
	strcpy( stats.cputime, "00:00:00" );
	strcpy( stats.timeleft, "00:00:00" );
	stats.nwus = 0;
	stats.nresults = 0;
	strcpy( stats.totalcpu, "0/000" );

	stats.pid = get_pid(dir);
	if ( stats.pid )
		stats.status = get_status(stats.pid);

        read_workunit_stats(dir);
        read_user_stats(dir);
}

/******************************************/
/* Read client pid                        */
/******************************************/

pid_t get_pid(const char *dir)
{
	pid_t pid = 0;
	char buf[128];
	FILE *fp;
	
	sprintf(buf, "%s/%s", dir, FILE_PID);

	if ( fp = fopen( buf, "r" ) ) {
		while ( fgets( buf, sizeof(buf), fp ) ) {
			pid = atoi(buf);
		}
		fclose(fp);
	} else if ( debug )
		fprintf(stderr, "INF: Failed to open pid file - %s\n", buf);

	return pid; 
}

/******************************************/
/* Read workunit statistics               */
/******************************************/

void read_workunit_stats(const char *dir)
{
	int search = 2;
	double timeleft = 0, timedone = 0, progress = 0;
	char buf[256];
	char *right;
	FILE *fp;

	sprintf(buf, "%s/%s", dir, FILE_STATE);

	if ( fp = fopen( buf, "r" ) ) {
		while ( fgets( buf, sizeof(buf), fp ) && search ) {
			if ( right = strchr( buf, '=' ) ) {
				*right++ = '\0';
				if ( ! strcmp( buf, "cpu" ) ) {
					timedone = strtod( right, NULL );
					secs_2_time( timedone, stats.cputime );

					search--;
				} else if ( ! strcmp( buf, "prog" ) ) {
					progress = strtod( right, NULL );
					sprintf(stats.progress, "%.2f%%", progress * 100);

					search--;
				}
			}
		}
		fclose(fp);

		if ( progress ) {
			timeleft = timedone/progress - timedone;
			secs_2_time( timeleft, stats.timeleft );
		}

		sprintf(buf, "%s/%s", dir, FILE_RESULT);
		if ( fp = fopen( buf, "r" ) ) {
			strcpy ( stats.timeleft, "UPLOAD" );
			fclose(fp);
		}
	} else {
		sprintf(buf, "%s/%s", dir, FILE_WTEMP);
		if ( fp = fopen( buf, "r" ) ) {
			strcpy( stats.timeleft, "DOWNLOAD" );
			fclose(fp);
		} else {
			strcpy( stats.timeleft, "CONNECT" );
			if ( debug )
				fprintf(stderr, "INF: Failed to open workunit file - %s/%s\n", dir, FILE_STATE);
		}
	}
}

/******************************************/
/* Read user statistics                   */
/******************************************/

void read_user_stats(const char *dir)
{
	char buf[256];
	char *right;
	FILE *fp;

	sprintf(buf, "%s/%s", dir, FILE_USERINFO);

	if ( fp = fopen( buf, "r" ) ) {
		while ( fgets( buf, sizeof(buf), fp ) ) {
			if ( right = strchr( buf, '=' ) ) {
				*right++ = '\0';
				if ( ! strcmp( buf, "nwus" ) )
					stats.nwus = atoi( right );
				else if ( ! strcmp( buf, "nresults" ) )
					stats.nresults = atoi( right );
				else if ( ! strcmp( buf, "total_cpu" ) )
					secs_2_epoch( atoi( right ), stats.totalcpu );
			}
		}
		fclose(fp);
	} else if ( debug )
		fprintf(stderr, "INF: Failed to open user file - %s\n", buf);
}

/******************************************/
/* Convert seconds to hh:mm:ss            */
/******************************************/

char *secs_2_time(const double seconds, char *time)
{
	int h = 0, m = 0, s = 0;
	double secs = seconds;

	strcpy(time, "99:99:99"); 

	if ( secs < 0 )
		return time;
	else if ( secs > 3600 ) {
		h = secs / 3600;
		secs -= h * 3600;
	}

	m = secs / 60;
	secs -= m * 60;
	s = secs;

	sprintf(time, "%02d:%02d:%02d", h, m, s);

	return time;
}

/******************************************/
/* Convert seconds to yy:ddd or ddd:hh    */
/******************************************/

char *secs_2_epoch(const double seconds, char *epoch)
{
	int a = 0, b = 0;
	double secs = seconds;

	strcpy( epoch, "0/000" ); 

	if ( secs < 3600 )
		return epoch;
	else if ( secs < 31536000 ) {
		a = secs / 86400;
		secs -= a * 86400;
		b = secs / 3600;
		sprintf(epoch, "%d/%02d", a, b);
	} else {
		a = secs / 31536000;
		secs -= a * 31536000;
		b = secs / 86400;
		sprintf(epoch, "%d/%03d", a, b);
	}

	return epoch;
}

/******************************************/
/* Read status of process                 */
/******************************************/

char get_status(const pid_t pid)
{
#ifdef __FreeBSD__
	char search[10][16];
#endif
#ifdef	__sun__
	psinfo_t ps;
#endif
	int i;
	char dest = 'U';
	char buf[128];
	FILE *fp;

	if ( pid == 0 )
		return dest;

#ifdef __FreeBSD__
	/* This is the non-advised method from Unix Programmers	*/
	/* FAQ. Could a FreeBSD user please submit a better	*/
	/* method :-)						*/

	sprintf(buf, "ps -lp %d 2> /dev/null", pid);

	if ( fp = popen( buf, "r" ) ) {
		fgets( buf, sizeof(buf), fp );
		sscanf( buf, "%s %s %s %s %s %s %s %s %s %s", &search[0], &search[1],&search[2], &search[3],
			&search[4], &search[5],&search[6], &search[7],&search[8], &search[9] );
		for ( i = 0; i < 10; i++ ) {
			if ( strstr( search[i], "STAT" ) )
				break;
		}
		if ( i < 10 ) {	
			fgets( buf, sizeof(buf), fp );
			if ( strstr( buf, SETI_PROG ) ) {
				sscanf( buf, "%s %s %s %s %s %s %s %s %s %s", &search[0], &search[1],
					&search[2], &search[3], &search[4], &search[5],
					&search[6], &search[7], &search[8], &search[9] );
				sprintf(dest, "%c", search[i][0]);
			}
		}
		fclose(fp);
	}
#endif /* __FreeBSD__ */

#ifdef __sun__
	sprintf(buf, "/proc/%d/psinfo", pid);

	if ( fp = fopen( buf, "r" ) ) {
		if ( read(fileno(fp), &ps, sizeof(psinfo_t)) != -1 ) {
			if ( strstr( ps.pr_fname, SETI_PROG ) )
				dest = ps.pr_lwp.pr_sname;
		}
		fclose(fp);
	}
#endif /* __sun__ */

#ifdef __linux__
	sprintf(buf, "/proc/%d/status", pid);

	if ( fp = fopen( buf, "r" ) ) {
		fgets( buf, sizeof(buf), fp );
		if ( strstr( buf, SETI_PROG ) ) {
			fgets( buf, sizeof(buf), fp );
			sscanf( buf, "%*s %c", &dest );
		}
		fclose(fp);
	}
#endif /* __linux__ */

	return dest;
}
